"""Auth providers for the Layer SDK."""

import os
import time
from abc import ABC, abstractmethod
from typing import Dict, Optional

import orjson
import requests

from .logger import logger
from .exceptions import (
    LayerAuthError,
    LayerHTTPError,
    LayerRequestError,
    LayerMissingRequiredConfigurationError,
)
from ._default_client import _create_default_http_session


class AuthProvider(ABC):
    """Abstract class for authentication providers.

    Implementations should provide a method to get the headers to be used in the request.
    """

    @abstractmethod
    def get_headers(self) -> Dict[str, str]:
        """Get the headers to be used in the request.

        Returns:
            Dict[str, str]: A dictionary of headers
        """
        pass


class OIDCClientCredentials(AuthProvider):
    """OIDC Client Credentials authentication provider.

    This provider uses the client credentials flow to obtain an access token.
    It is tested with Keycloak but can work with more providers.

    Args:
        token_url (str): The token URL to use to obtain the access token
        client_id (str): The client ID to use to obtain the access token
        client_secret (str): The client secret to use to obtain the access token
        scope (str): The scope to use to obtain the access token
        http_session (requests.Session): The HTTP client to use to make requests
        http_timeout (int): The timeout to use for the HTTP requests
    """

    _client_id: str
    _client_secret: str
    _token_url: str
    _access_token: Optional[str]
    _expires_at: float
    _token_type: str
    _scope: Optional[str]
    _http_session: requests.Session
    _http_timeout: int

    def __init__(
        self,
        *,
        token_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        scope: Optional[str] = None,
        http_session: Optional[requests.Session] = None,
        http_timeout: Optional[int] = None,
    ):
        """Initialize the OIDC Client Credentials authentication provider.

        Args:
            token_url (str): The token URL to use to obtain the access token
                e.g. https://keycloak.example.com/auth/realms/myrealm/protocol/openid-connect/token
            client_id (str): The client ID to use to obtain the access token
            client_secret (str): The client secret to use to obtain the access token
            scope (str): The scope to use to obtain the access token e.g. layer-sdk
            http_session (requests.Session): The HTTP client to use to make requests.
                By default, it uses a pre-configured session with retries and pool.
            http_timeout (int): The timeout to use for the HTTP requests.
                By default, it uses 10 seconds.

        Raises:
            LayerMissingRequiredConfigurationError: If any of the required arguments are missing
        """
        token_url = token_url or os.getenv("LAYER_OIDC_TOKEN_URL")
        if not token_url:
            raise LayerMissingRequiredConfigurationError("Missing required argument `token_url`")
        self._token_url = token_url

        client_id = client_id or os.getenv("LAYER_OIDC_CLIENT_ID")
        if not client_id:
            raise LayerMissingRequiredConfigurationError("Missing required argument `client_id`")
        self._client_id = client_id

        client_secret = client_secret or os.getenv("LAYER_OIDC_CLIENT_SECRET")
        if not client_secret:
            raise LayerMissingRequiredConfigurationError(
                "Missing required argument `client_secret`"
            )
        self._client_secret = client_secret

        self._scope = scope or os.getenv("LAYER_OIDC_SCOPE")
        self._access_token = None
        self._expires_at = 0.0
        self._token_type = "Bearer"

        if not http_session:
            http_session = _create_default_http_session()
        self._http_session = http_session

        self._http_timeout = http_timeout or int(os.getenv("LAYER_OIDC_TIMEOUT", 10))

        logger.debug(
            f"OIDCClientCredentials initialized with "
            f"token_url={self._token_url}, scope={self._scope}"
        )

    def _refresh_token(self):
        """Refresh the access token.

        Raises:
            LayerHTTPError: If the request to refresh the token fails
        """
        data = {
            "grant_type": "client_credentials",
            "client_id": self._client_id,
            "client_secret": self._client_secret,
        }

        if self._scope:
            data["scope"] = self._scope

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }

        try:
            response = self._http_session.post(
                self._token_url, data=data, headers=headers, timeout=self._http_timeout
            )
        except requests.RequestException as e:
            logger.error(f"Failed request to refresh token: {e}")
            raise LayerRequestError(f"Failed request to refresh token: {e}")

        if response.status_code == 401 or response.status_code == 403:
            logger.error(
                f"Failed to refresh token: " 
                f"{response.status_code} {response.content.decode('utf-8')}"
            )
            raise LayerAuthError(
                f"Failed to refresh token {response.status_code}: "
                f"{response.content.decode('utf-8')}"
            )

        if response.status_code != 200:
            logger.error(
                f"Failed request to refresh token: "
                f"{response.status_code} {response.content.decode('utf-8')}"
            )
            raise LayerHTTPError(
                "Failed request refresh token",
                response.status_code,
                response.content.decode("utf-8"),
            )

        token_data = orjson.loads(response.content)
        self._access_token = token_data["access_token"]
        self._token_type = token_data["token_type"] or self._token_type

        if token_data.get("expires_in"):
            logger.debug(f"Token refreshed, expires in {token_data['expires_in']} seconds")
            self._expires_at = (
                time.time() + token_data["expires_in"] - 60
            )  # Refresh 1 minute before expiration

    def get_headers(self) -> Dict[str, str]:
        """Get the headers to be used in the request.

        Raises:
            LayerHTTPError: If the request to refresh the token fails

        Returns:
            Dict[str, str]: A dictionary of headers
        """
        if 0 < self._expires_at < time.time() or not self._access_token:
            logger.debug("Token expired or not set, refreshing")
            self._refresh_token()

        return {"Authorization": f"{self._token_type} {self._access_token}"}



"""The main entry point for the Layer SDK i.e. the client."""

import os
import json
import uuid
import inspect
from typing import Any, Dict, List, Union, Optional
from datetime import datetime, timezone
from functools import wraps

import orjson
import requests

from .auth import AuthProvider
from .logger import logger
from .schemas import (
    SessionRequest,
    SessionActionKind,
    SessionActionRequest,
    FirewallLookupDecision,
    FirewallSessionLookupResponse,
)
from ._version import __version__
from ._platform import Platform, get_platform, platform_session_attributes
from .exceptions import (
    LayerHTTPError,
    LayerRequestError,
    LayerFirewallException,
    LayerFirewallHTTPError,
    LayerFirewallRequestError,
    LayerFirewallSessionBlocked,
    LayerRequestPreparationError,
    LayerMissingRequiredConfigurationError,
)
from ._default_client import _create_default_http_session
from .instrumentation import instrumentor_instances


class Client:
    """The main entry point for the Layer SDK i.e. the client.

    The client is a singleton class that is responsible
    for initializing the SDK and creating sessions and actions.
    """

    # Singleton instance
    _instance = None
    # Initialization flag
    _initialized: bool
    # Application ID
    _application_id: Union[str, uuid.UUID]
    # Base URL for the Layer API
    _base_url: str
    # Environment
    _environment: Optional[str]
    # Platform
    _platform: Platform
    # Auth provider
    _auth_provider: Optional[AuthProvider]
    # HTTP client session
    _http_session: requests.Session
    # HTTP timeout
    _http_timeout: int
    # Disabled instrumentors
    _disabled_instrumentors: List[str]
    # Firewall Base URL
    _firewall_base_url: Optional[str]
    # Enable firewall instrumentation
    _enable_firewall_instrumentation: bool

    def __new__(cls):
        """Singleton pattern to ensure only one instance of the SDK is created."""
        if cls._instance is None:
            cls._instance = super(Client, cls).__new__(cls)
            cls._instance._initialized = False

        return cls._instance

    def init(
        self,
        *,
        base_url: Optional[str] = None,
        application_id: Optional[Union[str, uuid.UUID]] = None,
        environment: Optional[str] = None,
        auth_provider: Optional[AuthProvider] = None,
        http_session: Optional[requests.Session] = None,
        http_timeout: Optional[int] = None,
        platform: Optional[Platform] = None,
        disabled_instrumentors: Optional[List[str]] = None,
        firewall_base_url: Optional[str] = None,
        enable_firewall_instrumentation: Optional[bool] = False,
    ):
        """Initialize the Layer SDK with the required configuration.

        Args:
            base_url (str, optional): The base URL for the Layer API.
                Defaults to the `LAYER_BASE_URL` environment variable.
            application_id (str, optional): The application ID.
                Defaults to the `LAYER_APPLICATION_ID` environment variable.
            environment (str, optional): The environment.
                Defaults to the `LAYER_ENVIRONMENT` environment variable.
            auth_provider (AuthProvider, optional): The authentication provider. Defaults to None.
            http_session (requests.Session, optional): The HTTP session/client.
                Defaults to preconfigured HTTP client with retry and pool.
            http_timeout (int, optional): The HTTP timeout.
                Defaults to `LAYER_HTTP_TIMEOUT` environment variable or 10 seconds
            platform (Platform, optional): The platform. Defaults to None.
            disabled_instrumentors (List[str], optional): The list of disabled instrumentors.
                Defaults to None i.e. enabling all.
            firewall_base_url (str, optional): The base URL for the firewall API.
            enable_firewall_instrumentation (bool, optional): Enable firewall instrumentation.
                It will automatically call firewall API for each session action creation.

        Raises:
            LayerMissingRequiredConfigurationError: If the required configuration is missing.
            LayerAlreadyInitializedError: If the SDK is already initialized.
        """
        if self._initialized:
            logger.error("Layer SDK already initialized")

            return

        base_url = base_url or os.environ.get("LAYER_BASE_URL")
        if not base_url:
            raise LayerMissingRequiredConfigurationError("Missing required argument `base_url`")

        self._base_url = base_url.strip().strip('"').strip("'").rstrip("/")

        if not platform:
            platform = get_platform()
        self._platform = platform

        environment = environment or os.environ.get("LAYER_ENVIRONMENT")
        if environment:
            environment = environment.lower()
        self._environment = environment

        application_id = application_id or os.environ.get("LAYER_APPLICATION_ID")
        if not application_id:
            raise LayerMissingRequiredConfigurationError("Application ID must be provided")
        self._application_id = application_id

        firewall_base_url = firewall_base_url or os.environ.get("LAYER_FIREWALL_BASE_URL")
        self._firewall_base_url = firewall_base_url

        self._auth_provider = auth_provider

        if not http_session:
            http_session = _create_default_http_session()
        self._http_session = http_session

        self._http_timeout = http_timeout or int(os.environ.get("LAYER_HTTP_TIMEOUT", 10))
        self._enable_firewall_instrumentation = enable_firewall_instrumentation or False
        if self._enable_firewall_instrumentation and not self._firewall_base_url:
            raise LayerMissingRequiredConfigurationError("Firewall base URL must be provided "
                                                         "if firewall instrumentation is enabled")

        self._disabled_instrumentors = disabled_instrumentors if disabled_instrumentors else []
        self._instrument()

        self._initialized = True
        logger.debug(
            f"Layer SDK initialized with base_url={self._base_url}, "
            f"application_id={self._application_id}, "
            f"environment={self._environment}, "
            f"firewall_base_url={self._firewall_base_url}"
        )

    def _instrument(self):
        """Instruments the SDK with the available instrumentors."""
        for name, instrumentor_class in instrumentor_instances.items():
            if name in self._disabled_instrumentors:
                logger.debug(f"Instrumentor {name} is disabled")

                continue

            instrumentor_instance = instrumentor_class(self)
            if not instrumentor_instance.supports():
                logger.warning(f"Instrumentor {name} is not supported")

                continue

            try:
                instrumentor_instance.instrument()
            except Exception as e:
                logger.error("Failed to instrument %s: %s", name, e)

                continue

            logger.debug(f"Instrumentor {name} is enabled")

    def _get_headers(self) -> Dict[str, str]:
        """Get the headers for the HTTP request.

        Returns:
            Dict[str, str]: The headers for the HTTP request
        """
        headers = {}
        if self._auth_provider:
            headers = self._auth_provider.get_headers()

        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        headers["User-Agent"] = f"layer-sdk-python/{__version__}"

        return headers

    def _encode_request(self, request: Any) -> bytes:
        try:
            body = orjson.dumps(
                request,
                option=orjson.OPT_UTC_Z
                | orjson.OPT_NAIVE_UTC
                | orjson.OPT_OMIT_MICROSECONDS
                | orjson.OPT_SERIALIZE_UUID
                | orjson.OPT_SERIALIZE_DATACLASS,
            )
        except TypeError as e:
            logger.warning(f"Failed to serialize request: {e}. Falling back to json.")

            try:
                body = json.dumps(
                    request,
                    ensure_ascii=True,
                ).encode("utf-8")
            except Exception as e:
                raise LayerRequestPreparationError(f"Failed to serialize request: {e}")

        return body

    def create_session(self, **kwargs) -> str:
        """Create a new session.

        Args:
            **kwargs: The session attributes. See `SessionRequest` for more information.

        Raises:
            LayerHTTPError: If the HTTP request fails.
            LayerRequestPreparationError: If the request preparation fails.

        Returns:
            str: The session ID
        """
        url = f"{self._base_url}/v1/sessions"

        try:
            request = SessionRequest(
                application_id=self._application_id,
                **kwargs,
            )
        except TypeError as e:
            raise LayerRequestPreparationError(str(e))

        if self._environment:
            request.attributes["environment"] = self._environment

        request.attributes = {
            **request.attributes,
            **platform_session_attributes(__version__, platform=self._platform),
        }

        try:
            response = self._http_session.post(
                url,
                data=self._encode_request(request),
                headers=self._get_headers(),
                timeout=self._http_timeout,
            )
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed request to create session: {e}")
            raise LayerRequestError(f"Failed request to create session: {e}")

        if response.status_code != 202:
            logger.error(
                f"Failed to create session: "
                f"{response.status_code} {response.content.decode('utf-8')}"
            )

            raise LayerHTTPError(
                "Failed to create session",
                response.status_code,
                response.content.decode("utf-8"),
            )

        session_id = orjson.loads(response.content)["session_id"]
        logger.debug(f"Session created with ID: {session_id}")

        return session_id

    def append_action(self, session_id: Union[str, uuid.UUID], **kwargs) -> None:
        """Append an action to a session.

        Args:
            session_id (Union[str, uuid.UUID]): The session ID
            **kwargs: The action attributes. See `SessionActionRequest` for more information.

        Raises:
            LayerRequestPreparationError: If the request preparation fails.
            LayerHTTPError: If the HTTP request fails.
            LayerFirewallSessionBlocked: If the firewall recommends to block the session.
                Only raised if `enable_firewall_instrumentation` is set to True.
        """
        if self._enable_firewall_instrumentation:
            try:
                firewall_lookup_response = self.firewall_session_lookup(session_id)

                if firewall_lookup_response.decision == FirewallLookupDecision.BLOCK:
                    raise LayerFirewallSessionBlocked(f"Session '{session_id}' is "
                                                      f"blocked by the firewall")

                logger.info(f"Session '{session_id}' was evaluated by the firewall. "
                            f"Decision: {firewall_lookup_response.decision}")
            except LayerFirewallException as e:
                logger.error(f"Failed to lookup session '{session_id}' in firewall: {e}. "
                             f"Skipping firewall check")

        url = f"{self._base_url}/v1/sessions/{str(session_id)}/actions"

        try:
            request = SessionActionRequest(**kwargs)
        except TypeError as e:
            raise LayerRequestPreparationError(str(e))

        try:
            response = self._http_session.post(
                url,
                data=self._encode_request(request),
                headers=self._get_headers(),
                timeout=self._http_timeout,
            )
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed request to append session action: {e}")
            raise LayerRequestError(f"Failed request to append session action: {e}")

        if response.status_code != 202:
            logger.error(
                f"Failed to append session action: "
                f"{response.status_code} {response.content.decode('utf-8')}"
            )
            raise LayerHTTPError(
                "Failed to append session action",
                response.status_code,
                response.content.decode("utf-8"),
            )

        logger.debug(f"Session action appended to session {session_id}")

    def firewall_session_lookup(
        self,
        session_id: Union[str, uuid.UUID]
    ) -> FirewallSessionLookupResponse:
        """Lookup a session in the firewall.

        Args:
            session_id (Union[str, uuid.UUID]): The session ID

        Raises:
            LayerFirewallRequestError: If the request to the firewall fails.
            LayerFirewallHTTPError: If the firewall returns an HTTP error.
            LayerMissingRequiredConfigurationError: If the firewall base URL is not provided.

        """
        if not self._firewall_base_url:
            raise LayerMissingRequiredConfigurationError("Firewall base URL must be provided")

        url = f"{self._firewall_base_url}/firewall/v1/lookup"

        try:
            response = self._http_session.post(
                url,
                data=self._encode_request({
                    "session_id": str(session_id),
                }),
                headers=self._get_headers(),
                timeout=self._http_timeout,
            )
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed request to the firewall session lookup: {e}")
            raise LayerFirewallRequestError(f"Failed request to the firewall session lookup: {e}")

        if response.status_code != 200:
            logger.error(
                f"Failed firewall session lookup: "
                f"{response.status_code} {response.content.decode('utf-8')}"
            )
            raise LayerFirewallHTTPError(
                "Failed firewall session lookup",
                response.status_code,
                response.content.decode("utf-8"),
            )

        return FirewallSessionLookupResponse(**orjson.loads(response.content))

    def session(
        self,
        *,
        session_id: Optional[Union[str, uuid.UUID]] = None,
        attributes: Optional[Dict[str, Any]] = None
    ):
        """Decorate to create a new session for a function.

        Args:
            session_id (Union[str, uuid.UUID], optional):
                Optional custom session ID if needs to be enforced.
            attributes (Dict[str, Any], optional): The session attributes.
                See `SessionRequest` for more information.

        Returns:
            Callable: The decorator
        """

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                new_session_id = self.create_session(attributes=attributes, session_id=session_id)

                sig = inspect.signature(func)
                if 'session_id' in sig.parameters:
                    kwargs['session_id'] = new_session_id

                result = func(*args, **kwargs)

                return result

            return wrapper

        return decorator

    def action(
        self,
        *,
        session_id: Union[str, uuid.UUID],
        kind: SessionActionKind,
        attributes: Optional[Dict[str, Any]] = None
    ):
        """Decorate to append an action to a session for a function.

        The wrapped function must return the tuple `(data, error, scanners)`.
        The `data` is the result of the action, the `error` is the error if any,
        and the `scanners` are the scanners if any.

        Args:
            session_id (Union[str, uuid.UUID]): The session ID
            kind (SessionActionKind): The action kind
            attributes (Dict, optional): The action attributes (see `SessionActionRequest`).

        Returns:
            Callable: The decorator
        """

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                start_time = datetime.now(timezone.utc)
                result = func(*args, **kwargs)
                end_time = datetime.now(timezone.utc)

                data, error, scanners = result, None, None
                if isinstance(result, tuple):
                    data, error, scanners = result + (None,) * (3 - len(result))

                self.append_action(
                    session_id=session_id,
                    kind=kind,
                    start_time=start_time,
                    end_time=end_time,
                    attributes=attributes or None,
                    data=data,
                    error=error,
                    scanners=scanners,
                )

                return result

            return wrapper

        return decorator



from typing import List, Optional
from datetime import datetime, timezone
from dataclasses import asdict

import wrapt

from ._base import BaseInstrumentor
from .._utils import get_module_version, is_module_version_more_than
from ..logger import logger
from ..schemas import (
    SessionActionKind,
    SessionActionError,
    SessionActionRequest,
    SessionActionScanner,
)


class OpenAIInstrumentation(BaseInstrumentor):
    """Instrumentation class for OpenAI SDK."""

    def supports(self) -> bool:
        """Check if the OpenAI SDK is supported."""
        version_correct = is_module_version_more_than("openai", "1.18.0")

        if version_correct is None:
            return False

        return version_correct

    def instrument(self):
        """Instrument the OpenAI SDK."""
        self._instrument_chat_completion()
        self._instrument_embedding()
        self._instrument_moderation()

    def _extract_session_id(self, kwargs: dict) -> Optional[str]:
        headers = kwargs.get("extra_headers", {})

        if "Layer-Session-Id" in headers:
            return headers.get("Layer-Session-Id")

        return headers.get("X-Layer-Session-Id")

    def _create_or_append(
            self,
            session_id: Optional[str],
            actions: List[SessionActionRequest],
    ) -> str:
        if session_id:
            for action in actions:
                self._layer.append_action(
                    session_id,
                    **asdict(action),
                )

                logger.debug(f"Appended {action.kind.value} action to '{session_id}' session")

            return session_id

        logger.debug(f"No session id was provided. "
                     f"Creating a new session with {len(actions)} actions")

        new_session_id = self._layer.create_session(
            attributes={
                "source": "instrumentation.openai",
                "openai.version": get_module_version("openai")
            },
            actions=actions,
        )

        return new_session_id

    def _process_completion_stream(self, session_id, start_time, result):
        completion_content = ""
        completion_role = ""
        completion_model = ""
        chunk_count = 0
        for chunk in result:
            if len(chunk.choices) > 0:
                if not hasattr(chunk.choices[0], "delta"):
                    continue

                if hasattr(chunk.choices[0].delta, "content"):
                    content = chunk.choices[0].delta.content
                    if content:
                        completion_content += content

                if hasattr(chunk.choices[0].delta, "role"):
                    role = chunk.choices[0].delta.role
                    if role:
                        completion_role = role

                completion_model = chunk.model
                chunk_count += 1

                yield chunk

        self._create_or_append(session_id, [
            SessionActionRequest(
                kind=SessionActionKind.COMPLETION_OUTPUT,
                start_time=start_time,
                end_time=datetime.now(timezone.utc),
                attributes={
                    "stream": "true",
                    "model.id": completion_model,
                    "chunk.count": chunk_count,
                },
                data={
                    "messages": [
                        {
                            "content": completion_content,
                            "role": completion_role,
                        }
                    ],
                },
            )
        ])

    def _instrument_chat_completion(self):
        @wrapt.patch_function_wrapper('openai.resources.chat.completions', 'Completions.create')
        def wrapper(wrapped, instance, args, kwargs):
            messages = [
                {
                    "role": message.get("role"),
                    "content": message.get("content"),
                    "user": message.get("user", ""),
                }
                for message in kwargs.get("messages")
            ]

            attrs = {
                "model.id": kwargs.get("model"),
                "model.stream": "true" if kwargs.get("stream", False) else "false",
            }

            attr_has_keys = [
                "response_format",
                "metadata",
                "tools",
            ]
            for key in attr_has_keys:
                attrs[f"model.has_{key}"] = "false" if kwargs.get(key, None) else "true"

            attr_keys = [
                "max_tokens",
                "temperature",
                "frequency_penalty",
                "max_completion_tokens",
                "logit_bias",
                "logprobs",
                "modalities",
                "n",
                "presence_penalty",
                "stop",
                "seed",
                "top_p",
                "user",
            ]
            for key in attr_keys:
                if key in kwargs and kwargs[key] is not None:
                    attrs[f"model.{key}"] = kwargs[key]

            session_id = self._create_or_append(self._extract_session_id(kwargs), [
                SessionActionRequest(
                    kind=SessionActionKind.COMPLETION_PROMPT,
                    start_time=datetime.now(timezone.utc),
                    end_time=datetime.now(timezone.utc),
                    attributes=attrs,
                    data={
                        "messages": messages,
                    },
                )
            ])

            result = None
            exception_to_raise = None
            start_time = datetime.now(timezone.utc)
            try:
                result = wrapped(*args, **kwargs)
            except Exception as e:
                exception_to_raise = e

            if exception_to_raise or result is None:
                self._create_or_append(session_id, [
                    SessionActionRequest(
                        kind=SessionActionKind.COMPLETION_OUTPUT,
                        start_time=start_time,
                        end_time=datetime.now(timezone.utc),
                        error=SessionActionError(message=str(exception_to_raise))
                    )
                ])

                raise exception_to_raise

            is_stream = kwargs.get("stream", False)
            if is_stream:
                return self._process_completion_stream(session_id, start_time, result)

            completions_scanners = None
            completions_messages = []

            if result.usage is not None:
                completions_scanners = [
                    SessionActionScanner(
                        name="usage",
                        data={
                            "prompt_tokens": result.usage.prompt_tokens,
                            "completion_tokens": result.usage.completion_tokens,
                            "total_tokens": result.usage.total_tokens,
                        },
                    )
                ]

            for choice in result.choices:
                if choice.message.content is None:
                    continue

                completions_messages.append({
                    "content": choice.message.content,
                    "role": choice.message.role,
                })

            self._create_or_append(session_id, [
                SessionActionRequest(
                    kind=SessionActionKind.COMPLETION_OUTPUT,
                    start_time=start_time,
                    end_time=datetime.now(timezone.utc),
                    attributes={
                        "model.created": result.created,
                        "model.id": result.model,
                        "model.completion_id": result.id,
                    },
                    data={
                        "messages": completions_messages,
                    },
                    scanners=completions_scanners,
                )
            ])

            return result

    def _instrument_embedding(self):
        @wrapt.patch_function_wrapper('openai.resources.embeddings', 'Embeddings.create')
        def wrapped_embedding_create(wrapped, instance, args, kwargs):
            error = None
            exception_to_raise = None
            result = None

            start_time = datetime.now(timezone.utc)
            try:
                result = wrapped(*args, **kwargs)  # type: ignore
            except Exception as e:
                exception_to_raise = e
                error = SessionActionError(message=str(e))
            finally:
                end_time = datetime.now(timezone.utc)

            scanners = None
            if error is None:
                scanners = [
                    SessionActionScanner(
                        name="usage",
                        data={
                            "prompt_tokens": result.usage.prompt_tokens,  # type: ignore[reportOptionalMemberAccess]
                            "total_tokens": result.usage.total_tokens,  # type: ignore[reportOptionalMemberAccess]
                        },
                    )
                ]

            embedding_input = kwargs.get("input")
            if isinstance(embedding_input, str):
                embedding_input = [embedding_input]

            attributes = {
                "model.id": kwargs.get("model", "text-embedding-ada-002"),
                "model.encoding_format": kwargs.get("encoding_format", "float"),
            }

            if "user" in kwargs:
                attributes["model.user"] = kwargs.get("user")

            action = SessionActionRequest(
                kind=SessionActionKind.EMBEDDINGS,
                start_time=start_time,
                end_time=end_time,
                attributes=attributes,
                data={
                    "input": embedding_input,
                },
                scanners=scanners,
                error=error
            )

            self._create_or_append(self._extract_session_id(kwargs), [action])

            if exception_to_raise:
                raise exception_to_raise

            return result

    def _instrument_moderation(self):
        @wrapt.patch_function_wrapper('openai.resources.moderations', 'Moderations.create')
        def wrapped_moderation_create(wrapped, instance, args, kwargs):
            error = None
            exception_to_raise = None
            result = None

            start_time = datetime.now(timezone.utc)
            try:
                result = wrapped(*args, **kwargs)
            except Exception as e:
                exception_to_raise = e
                error = SessionActionError(message=str(e))
            finally:
                end_time = datetime.now(timezone.utc)

            scanners = None
            if error is None and len(result.results) > 0:  # type: ignore[reportOptionalMemberAccess]
                scanners = []
                for index, item in enumerate(result.results):  # type: ignore[reportOptionalMemberAccess]
                    scanners.append(SessionActionScanner(
                        name=f"openai_moderation_{index}",
                        data=item.categories.to_dict(),
                    ))

            moderation_input = kwargs.get("input")
            if isinstance(moderation_input, str):
                moderation_input = [moderation_input]

            action = SessionActionRequest(
                kind=SessionActionKind.MODERATION,
                start_time=start_time,
                end_time=end_time,
                attributes={
                    "model.id": kwargs.get("model", "text-moderation-latest"),
                },
                data={
                    "input": moderation_input,
                },
                scanners=scanners,
                error=error
            )

            self._create_or_append(self._extract_session_id(kwargs), [action])

            if exception_to_raise:
                raise exception_to_raise

            return result



import os
import platform
from typing import Dict, Union, Literal, Optional
from functools import lru_cache


class OtherPlatform:
    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return f"Other:{self.name}"


Platform = Union[
    OtherPlatform,
    Literal[
        "MacOS",
        "Linux",
        "Windows",
        "iOS",
        "Android",
        "Unknown",
    ],
]


def get_platform() -> Platform:
    try:
        system = platform.system().lower()
        platform_name = platform.platform().lower()
    except Exception:
        return "Unknown"

    if "iphone" in platform_name or "ipad" in platform_name:
        # Tested using Python3IDE on an iPhone 11 and Pythonista on an iPad 7
        # system is Darwin and platform_name is a string like:
        # - Darwin-21.6.0-iPhone12,1-64bit
        # - Darwin-21.6.0-iPad7,11-64bit
        return "iOS"

    if system == "darwin":
        return "MacOS"

    if system == "windows":
        return "Windows"

    if "android" in platform_name:
        # Tested using Pydroid 3
        return "Android"

    if system == "linux":
        return "Linux"

    if platform_name:
        return OtherPlatform(platform_name)

    return "Unknown"


@lru_cache(maxsize=None)
def platform_session_attributes(version: str, *, platform: Optional[Platform]) -> Dict[str, str]:
    """Return a dictionary with the session attributes.

    They provide information about the platform and the SDK version.

    Args:
        version: The version of the SDK.
        platform: The platform where the SDK is running. If not provided, it will be detected.

    Returns:
        A dictionary with the session attributes.
    """
    default_attributes = {
        "source": f"layer-python-sdk/{version}",
        "os": str(platform or get_platform()),
        "python.runtime": _get_python_runtime(),
        "python.runtime.version": _get_python_version(),
    }

    return {**default_attributes, **kubernetes_session_attributes()}


def kubernetes_session_attributes() -> Dict[str, str]:
    if not os.environ.get('KUBERNETES_SERVICE_HOST'):
        return {}

    kubernetes_info = {
        "kubernetes_pod_name": os.environ.get(
            'POD_NAME',
            os.environ.get('HOSTNAME')
        ),  # POD_NAME or fallback to HOSTNAME
        "kubernetes_pod_namespace": os.environ.get('POD_NAMESPACE'),
        "kubernetes_node_name": os.environ.get('NODE_NAME'),
    }

    return {k: v for k, v in kubernetes_info.items() if v}  # Filter out None values



def _get_python_runtime() -> str:
    try:
        return platform.python_implementation()
    except Exception:
        return "unknown"


def _get_python_version() -> str:
    try:
        return platform.python_version()
    except Exception:
        return "unknown"



"""This module contains the schemas for the Layer SDK."""
import uuid
from enum import Enum
from typing import Any, Dict, List, Union, Optional
from datetime import datetime
from dataclasses import field, dataclass

from .exceptions import LayerSchemaValidationError


class SessionActionKind(Enum):
    """Enum for session action kinds.

    Attributes:
        EMBEDDINGS (str): Embeddings action
        MODERATION (str): Moderation action
        COMPLETION_PROMPT (str): Completion prompt action
        COMPLETION_OUTPUT (str): Completion output action
    """

    EMBEDDINGS = "embeddings"
    MODERATION = "moderation"
    COMPLETION_PROMPT = "completion_prompt"
    COMPLETION_OUTPUT = "completion_output"


@dataclass
class SessionActionError:
    """Error schema for session actions.

    Attributes:
        message (str): The error message. It is a required field
        code (int, optional): The error code
        details (Dict[str, Any], optional): The error details
    """

    message: str
    code: Optional[int] = None
    details: Optional[Dict[str, Any]] = None


@dataclass
class SessionActionScanner:
    """Scanner schema for session actions.

    Attributes:
        name (str): The scanner name
        data (Dict[str, Any]): The scanner data
    """

    name: str
    data: Dict[str, Any]


@dataclass
class SessionActionRequest:
    """Request schema for appending a new session action to an existing session.

    Attributes:
        kind (SessionActionKind): The action kind
        start_time (datetime): The action start time
        end_time (datetime): The action end time
        attributes (Dict[str, Any], optional): The action attributes
        data (Union[Dict, List], optional): The action data
        error (SessionActionError, optional): The action error
        scanners (List[Dict], optional): The action scanners
    """

    # The action kind. It is required field
    kind: SessionActionKind
    # The action start time. It is required field
    start_time: datetime
    # The action end time. It is required field
    end_time: datetime
    # The action attributes (i.e. metadata)
    attributes: Dict[str, Any] = field(default_factory=dict)
    # The action data (i.e. the result of the action)
    data: Optional[Dict] = None
    # The action error, if any
    error: Optional[SessionActionError] = None
    # The action scanners, if any
    scanners: Optional[List[SessionActionScanner]] = None


@dataclass
class SessionRequest:
    """Request schema for creating a new session.

    Attributes:
        application_id (Union[str, uuid.UUID]): The application ID
        session_id (Union[str, uuid.UUID], optional): The custom session ID
        attributes (Dict[str, str], optional): The session attributes
        actions (List[SessionActionRequest], optional): The session actions
    """

    # The application ID
    application_id: Union[str, uuid.UUID]
    # Custom session ID (UUID v5). If not provided, a new UUID will be generated
    session_id: Optional[Union[str, uuid.UUID]] = field(default=None)
    # The session attributes (i.e. metadata)
    attributes: Dict[str, str] = field(default_factory=dict)
    # The session actions, if any. When empty, the session will be created without any actions
    actions: Optional[List[SessionActionRequest]] = None

    def __post_init__(self):
        """Post initialization method to validate the schema."""
        if self.session_id is not None:
            if isinstance(self.session_id, uuid.UUID):
                session_id_uuid = self.session_id
            else:
                try:
                    session_id_uuid = uuid.UUID(self.session_id)
                except Exception as e:
                    raise LayerSchemaValidationError(e)

            if session_id_uuid.version != 5:
                raise LayerSchemaValidationError("Custom session ID must be a UUID v5")


@dataclass
class FirewallLookupResponseContextViolatedPolicy:
    """Violated policy schema for the firewall lookup response context.

    Attributes:
        id (int): The policy ID
        name (str): The policy name
        severity (str): The policy severity
        action (str): The policy action
    """
    id: int
    name: str
    severity: str
    action: str


@dataclass
class FirewallLookupResponseContext:
    """Context schema for the firewall lookup response.

    Attributes:
        violated_policies (List[FirewallLookupResponseContextViolatedPolicy], optional): Violated.
    """
    violated_policies: Optional[List[FirewallLookupResponseContextViolatedPolicy]] = None


class FirewallLookupDecision(Enum):
    """Enum for firewall lookup decisions.

    Attributes:
        NO_DATA (str): No data available yet
        PASS (str): No violations
        SKIP (str): Violations but only "monitor" actions
        ALERT (str): Violations but at most "inform" actions
        BLOCK (str): Violations with at least one "block" action
    """
    NO_DATA = "no_data"
    PASS = "pass"
    SKIP = "skip"
    ALERT = "alert"
    BLOCK = "block"


@dataclass
class FirewallSessionLookupResponse:
    """Response schema for the firewall session lookup.

    Attributes:
        session_id (uuid.UUID): The session ID
        decision (str): The firewall decision
        context (FirewallLookupResponseContext): The firewall context
    """

    session_id: uuid.UUID
    decision: FirewallLookupDecision
    context: FirewallLookupResponseContext



from abc import ABC, abstractmethod
from typing import Any


class BaseInstrumentor(ABC):
    """Base class for all instrumentors."""
    _layer: Any

    def __init__(self, layer: Any):
        """Initialize the instrumentor.

        Args:
            layer (Any): The Layer SDK instance.
        """
        self._layer = layer

    @abstractmethod
    def instrument(self):
        """Instrument the code."""
        pass

    @abstractmethod
    def supports(self) -> bool:
        """Check if the instrumentor supports the current environment."""
        pass



"""The logger module for the Layer SDK."""

import sys
import logging


def setup_logger(name: str, level: int = logging.DEBUG) -> logging.Logger:
    """Set up the logger.

    Args:
        name (str): The logger name
        level (int, optional): The logging level. Defaults to logging.INFO.

    Returns:
        logging.Logger: The logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)

    handler = logging.StreamHandler(sys.stdout)

    # e.g. [2023-10-05 14:12:26 - layer_minus_sdk._base_client:818 - DEBUG]
    # HTTP Request: POST http://127.0.0.1:4010/foo/bar "200 OK"
    formatter = logging.Formatter(
        "[%(asctime)s - %(name)s:%(lineno)d - %(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(handler)

    return logger


logger: logging.Logger = setup_logger("layer-sdk")
_urllib3_logger: logging.Logger = setup_logger("urllib3.connectionpool")



"""Instrumentation module for the SDK."""
from typing import Dict, Callable

from ._base import BaseInstrumentor
from ._openai import OpenAIInstrumentation

instrumentor_instances: Dict[str, Callable] = {
    'openai': OpenAIInstrumentation
}

__all__ = ["BaseInstrumentor", "instrumentor_instances"]



"""Exceptions for the Layer SDK."""

from __future__ import annotations


class LayerSDKException(Exception):
    """Base class for exceptions in this module."""


class LayerRequestError(LayerSDKException):
    """Exception raised for errors in the request."""


class LayerHTTPError(LayerSDKException):
    """Exception raised for errors in the HTTP request."""

    def __init__(self, message: str, status_code: int, response_data: str):
        """Initialize the exception.

        Args:
            message (str): The error message
            status_code (int): The HTTP status code
            response_data (str): The response data
        """
        self.status_code = status_code
        self.message = message
        self.response_data = response_data

        super().__init__(f"HTTP Error {self.status_code}: {self.message}\n{self.response_data}")


class LayerAuthError(LayerSDKException):
    """Exception raised when authentication fails."""


class LayerRequestPreparationError(LayerSDKException):
    """Exception raised when a request fails."""


class LayerMissingRequiredConfigurationError(LayerSDKException):
    """Exception raised when a required configuration is missing."""


class LayerSchemaValidationError(LayerSDKException):
    """Exception raised when a schema validation fails."""


class LayerFirewallException(LayerSDKException):
    """Exception raised for errors in the firewall."""


class LayerFirewallRequestError(LayerFirewallException):
    """Exception raised for errors in the request to the firewall."""


class LayerFirewallHTTPError(LayerHTTPError, LayerFirewallException):
    """Exception raised when a firewall returns an HTTP error."""


class LayerFirewallSessionBlocked(LayerSDKException):
    """Exception raised when a firewall recommends to block the session."""



import importlib.metadata
from typing import Optional

from packaging import version


def is_module_version_more_than(module_name: str, version_spec: str) -> Optional[bool]:
    """Check if the installed version of a module is less than the specified version.

    Parameters:
        module_name (str): The name of the module to check.
        version_spec (str): The version to compare against in the format
        'major.minor.patch'.

    Returns:
        bool: True if the installed version is less than the specified version,
        False otherwise.
    """
    installed_version = get_module_version(module_name)
    if not installed_version:
        return None

    return version.parse(installed_version) >= version.parse(version_spec)


def get_module_version(module_name: str) -> Optional[str]:
    """Get the installed version of a module.

    Parameters:
        module_name (str): The name of the module to check.

    Returns:
        str: The installed version of the module.
    """
    try:
        return importlib.metadata.version(module_name)
    except importlib.metadata.PackageNotFoundError:
        return None



"""Init file for the Layer SDK."""

from .auth import AuthProvider, OIDCClientCredentials
from .client import Client
from .schemas import (
    SessionRequest,
    SessionActionKind,
    SessionActionError,
    SessionActionRequest,
    SessionActionScanner,
    FirewallLookupDecision,
    FirewallSessionLookupResponse,
)
from .exceptions import (
    LayerAuthError,
    LayerHTTPError,
    LayerRequestError,
    LayerSDKException,
    LayerFirewallHTTPError,
    LayerFirewallRequestError,
    LayerSchemaValidationError,
    LayerFirewallSessionBlocked,
    LayerRequestPreparationError,
    LayerMissingRequiredConfigurationError,
)

layer = Client()

__all__ = [
    "layer",
    "AuthProvider",
    "SessionActionKind",
    "SessionRequest",
    "SessionActionError",
    "SessionActionRequest",
    "SessionActionScanner",
    "OIDCClientCredentials",
    "LayerAuthError",
    "LayerHTTPError",
    "LayerSDKException",
    "LayerMissingRequiredConfigurationError",
    "LayerRequestPreparationError",
    "LayerRequestError",
    "LayerFirewallSessionBlocked",
    "LayerFirewallHTTPError",
    "LayerSchemaValidationError",
    "LayerFirewallRequestError",
    "FirewallLookupDecision"
]



import requests
from requests.adapters import Retry, HTTPAdapter


def _create_default_http_session() -> requests.Session:
    """Create a default http client based on the requests with retry and timeout settings.

    We chose these settings to make sure
    that the client is robust and can handle most of the common cases.

    Returns:
        requests.Session: A session object with the default settings.
    """
    retry_strategy = Retry(
        total=3,
        status_forcelist=[502, 503, 504, 408, 425],
        backoff_factor=0.5,
        allowed_methods=None,
        raise_on_redirect=False,
        raise_on_status=False,
        respect_retry_after_header=True,
    )
    adapter = HTTPAdapter(
        max_retries=retry_strategy,
        pool_connections=3,
        pool_maxsize=10,
        pool_block=False,
    )
    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    return session
